#ifndef N_SPLINE_H
#define N_SPLINE_H

///@file
///  Defines cubic_spline a 3rd order spline.  Use spline instead.  

//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.
//
// Further warning, the algorithms container here are c++ transmogriphications
// of code from a college textbook -- originally fortran.  See below.
#include <stdio.h>
#include <map>
namespace cxxtls
{


class cubic_spline
  //
  ///  Cubic interpolatory spline.  Derived from the example in the book
  ///    Numerical Computing by Shampine and Allen.  Copyright 1973.
  ///    ISBN 0-8150-6.
  ///
  ///  Create the spline by copying, constructing, assigning etc.  Once you
  ///  have all the data points added (as above or using add_sample), you
  ///  can use the function call operator to evaluate the spline at any
  ///  X value.  For example:
  ///
  ///    cubic_spline s;  
  ///
  ///    s.add_sample(0,0);  // basically sinusoidal wave form
  ///    s.add_sample(1,1);
  ///    s.add_sample(2,0);
  ///    s.add_sample(3,-1);
  ///    s.add_sample(4,0);
  ///
  ///    double value = s(1.5);  // about .5
  ///
  ///  You can iterate over the actual sample data using the samples()
  ///  method.  You can interpolate multiple points using:
  ///
  ///     evaluate(1.0,.2, begin_output, end_output);
  ///
  ///  see below.
  ///
  ///  Note that once the coefficients have been calculated, you can
  ///  look at the sample data to get the 
{
public:

  struct info  ///  info stored for each x value in the sample data set
  {
    double y_; ///<  value of the function at the specified x location
    double s_; ///<  second derivative the interpolation function at x

    info(double y, double s=0)
    : y_(y),
      s_(s)
    {
    }
  };

  typedef std::multimap<double,info> rep_t;          ///<  sorted list of data pts
  typedef rep_t::value_type          value_t;        ///<  aids in inserting data
  typedef rep_t::iterator            iterator;       ///<  mutable iterator
  typedef rep_t::const_iterator      const_iterator; ///<  const iterator

  void add_sample(double x, double y)
      ///  Add a new data point to the sample and store it its proper
      ///  sorted order
  {
    dirty_ = true;

    iterator where = rep_.find(x);

    if(where != rep_.end())
      where->second.y_ = y;
    else
      rep_.insert(value_t(x,y));
    
  }

  void erase()
    ///  Eliminates all data points
  {
    rep_.erase(rep_.begin(), rep_.end());
    dirty_ = true;
  }

  double operator() (double x) const { return evaluate(x); }
    ///<  Evaluate the spline at location x.

  rep_t const& samples() const { return rep_; }
    ///<  Get a list of the actual sample data

  cubic_spline() ///   Default constructor
  : dirty_(true)
  {
  }

  cubic_spline(cubic_spline const& rhs) ///  Copy constructor
  : rep_(rhs.rep_),
    dirty_(rhs.dirty_)
  {
  }

  cubic_spline& operator=(cubic_spline const& rhs) ///  Assignment operator
  {
    this->cubic_spline::~cubic_spline();
    new(this) cubic_spline(rhs);
    return *this;
  }

  template<class InputIterator>
  cubic_spline(InputIterator first, InputIterator last)
  : dirty_(true)
  {
    ///<  construct from an iterator range.  The input data must refer to
    ///  pair<X,Y> where X and Y are assignment compatible with double.

    while(first != last)
    {
      add_sample(first->first, first->second);

      ++first;
    }
  }

  template<class OutputIterator>
  void evaluate(double start_x, 
		double delta_x, 
		OutputIterator begin,
		OutputIterator end
	       )
    //
    ///  Create a collection of interpolated values.  Output those
    ///  values through an iterator that points to a pair<X,Y> where 
    ///  X, and Y are assignment compatible with double.  The begin,end
    ///  variables define the standard output iterator range logic.  The
    ///  first value output will be (start_x, f(start_x)), the second will be
    ///  (start_x+delta_x, f(start_x+delta_x)), and so forth.
  {
     while(begin != end)
     {
       begin->first  = start_x;
       begin->second = evaluate(start_x);

       ++begin;

       start_x += delta_x;
     }
  }

   const_iterator ceil(double const& x) const 
     /// get an iterator to the sample whose X is immediately above 'x'.
     /// Returns end() if no such sample exists. O(ln(N)) performance
   { 
     return rep_.upper_bound(x); 
   } 

   const_iterator floor(double const x) const;
					    ///<
                                            /// get iterator to the sample
                                            /// whose X is less than or
                                            /// equal to x and which is higher
                                            /// than any other sample.
                                            /// O(ln(N)) performance

  double evaluate(double) const;            ///<  evaluate 1 point
  void   compute_coefficients();            ///<  compute s_'s.


private:

  rep_t rep_;
  bool  dirty_;

};
} // namespace cxxtls
#endif
