#ifndef str_h_included
#define str_h_included


//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.




///@file
///
///  This file defines string parsing and manipulation tools that are
///  commonly used but not part of the library.
//

#include <list>
#include <string>
#include <string.h>
#include <algorithm>
#include <vector>
#include <ctype.h>
#include <stdio.h>  // auto-included on some systems but not all


#ifdef _MSC_VER
#define STRNCASECMP strnicmp
#else
#define STRNCASECMP strncasecmp
#endif


namespace cxxtls
{

class StrTool
    /// This class just defines a bunch of string utility functions.
    /// They are part of a class just so that they don't clog up the
    /// global namespace.  I suppose a namespace would have worked
    /// just as well.
{
public:

  typedef std::string        string;
  typedef std::list<string>  stringlist_t;

  enum constants
  {
     all=2000000000  ///  two billion strings out to be enough to count as 'all'
  };

  struct Is_Space
    /// This struct is a 'functor' whose job it is to tell you whether a
    /// given character counts as 'space' in the normal way.  You can
    /// write your own that have more or less complex tests...  To use
    /// this struct, you construct an object of this type then execute the
    /// object like a funct -- passing it a character.  The return value from
    /// the function call is a boolean value that tells you whether or not
    /// the character you pass in is whitespace.
    ///
    /// This functor makes its decision based on what the character is.
    /// An alternative, is to decide based on what the character is not.
    /// For example, you could have a list of characters that are words
    /// and some functor could return true if the specified character were
    /// not in the list.
    ///
    /// To use Is_Space, declare a variable of type Is_Space and treat that
    /// variable as if it were a function accepting a character as a parameter
    /// and returning a flag indicating that the specified character is a "space".
    /// For example:
    ///
    /// @code
    ///    Is_Space is_space;
    ///
    ///    bool yes = is_space('\t');   // yep, tab is a space
    /// @endcode
  {
    bool operator() (char c) const
    {
      return c == ' '  ||  //  perform the comparison
             c == '\t' ||
             c == '\n' ||
             c == '\r' ;
    }

  };

  struct Is_Newline
    ///  If you want to use the parse_words method to split a string
    ///  into lines, this is the Is_Separator parameter type for you.
  {
    bool operator() (char c) const
    {
      return c == '\n';
    }
  };

  struct Is_Colon
    ///  If you want to use the parse_words method to split a string
    ///  into substrings, separated by colons, use this functor.
  {
    bool operator() (char c) const { return c == ':'; }
  };

  struct Is_Comma
    ///  If you want to use the parse_words method to split a string
    ///  into substrings, separated by commas, use this functor.
  {
    bool operator() (char c) const { return c == ','; }
  };

  struct Is_Tab
    ///  If you want to use the parse_words method to split a string
    ///  into substrings, separated by tabs, use this functor.
  {
    bool operator() (char c) const { return c == '\t'; }
  };

  struct Is_Delim
    ///  If you want to use the parse_words method to split a string
    ///  into substrings, separated by some delimiter, use this functor.
  {
    char delim_;

    Is_Delim(char c): delim_(c) {}

    bool operator() (char c) const { return c == delim_; }
  };

    static 
    size_t 
    splitOnString(std::string const &s, std::vector<std::string> &output, std::string const &sep)
       ///
       ///  Split a string, s, into pieces which are delimited by fixed separator string, sep.
       ///  the split pieces are appended to a vector.
       ///
       ///  Note that this function does not remove leading and trailing delimiters -- that
       ///  is leading and trailing instances of the sep string are not discarded.   Multiple
       ///  instances of the sep string are assumed to refer to separate output string values.
       ///  This is a "split" operation, not a parse operation.
       ///
       ///  @return the count of items added to the output
    {
        size_t count = 0;

        size_t pos=0;

        size_t size = s.size();

        size_t sepSize = sep.size();

        for(;;)
        {
           size_t newPos = s.find(sep, pos);

           if(newPos <= size)
           {
              // found a new one!

              output.push_back("");
              output.back() = s.substr(pos, newPos - pos);

              pos = newPos + sepSize;

              ++count;

           }
           else
           {
             output.push_back( s.c_str() + pos );

             ++count;

             break;
           }
        }


        return count;

    }

    template<class Is_Delim>
    static 
    size_t 
    splitOnDelimiter(std::string const &s, std::vector<std::string> &output, Is_Delim const &is_delim)
       ///
       ///  Split a string, s, into pieces which are delimited by any member of a fixed delimiter
       ///  set of characters.  The split pieces are APPENDED to a vector.
       ///
       ///  Note that this function does not remove leading and trailing delimiters -- that
       ///  is leading and trailing instances of the sep string are not discarded.   Multiple
       ///  instances of the sep string are assumed to refer to separate output string values.
       ///  This is a "split" operation, not a parse operation.
       ///
       ///  @return the count of items added to the output
    {
        size_t count = 0;

        std::string::const_iterator scan = s.begin();
        std::string::const_iterator end = s.end();

        while(scan != end)
        {
           std::string::const_iterator delimLoc = std::find_if(scan, end, is_delim);
           
           // found a new one!
           
           output.push_back(std::string());
           output.back().append(scan,delimLoc);
           
           scan = delimLoc;

           if(scan != end)
               ++scan;
           
           ++count;
           
        }


        return count;

    }




  template<class IsSeparator>
  static
  int  parse_words(string::const_iterator &cur,
                   string::const_iterator &end,
                   stringlist_t       *l,
                   unsigned            at_most,
                   IsSeparator const & is_space
                  )
    /// Parse a string, which is assumed to be collection of words
    /// separated by spaces.  The parsed words are stored in a list
    /// and the count of words is returned.  Note that leading and trailing
    /// spaces are removed.  This is a parse operation, not a split operation.
    /// If you want a split, rather than a parse, pick a different function.
    /// In this function, one or more spaces are considered to be "one space".
    ///
    /// You can specify a maximum limit of words to store in the list -- the 
    /// default is to store them all -- well 2 billion anyway.  If the maximum 
    /// limit is reached, all unparsed text is left in the last item in the
    /// list -- leading spaces will be removed from it, however.  That is,
    /// this function works much like the bourne shell 'read' command that
    /// reads words from a a file and stores them in environment variables.
    /// In this case, the parsed words are stored in a list of strings.
    ///
    /// The parameters are as follows:
    ///
    ///   @param s the string to be parsed
    ///
    ///   @param l a pointer to the list in which the parsed words will go
    ///            \note  words are appended to the end of the list, you will
    ///            have to clean it out of anything you don't want to appear
    ///            at the beginning of the list.
    ///
    ///   @param at_most the maximum number of elements to be parsed into
    ///                  the list, see note above about how unparsed text
    ///                  is handled.
    ///
    ///   @param is_space a functor describing how to tell if a character is
    ///                   a 'space' or not.  To make your own definitions, use
    ///                   StrTool::Is_Space as an example.
  {

    int count=0;

    while( at_most != 0  &&
           cur != end
         )
    {
      // parse a word and stick it in the list -- remove leading and
      // trailing spaces unless this is the 'last word' in which case
      // we remove leading but not trailing characters.

      while(cur != end && is_space(*cur) ) ++cur;

      if(cur == end )
          return count;

      if(at_most == 1)           // last string gets all of text
      {
         string word(cur,end);

         l->push_back(word);

         ++count;

         return count;
      }

      // strings other than the last have the trailing spaces removed
      // before insertion into the list

      string word;

      char c;

      while(cur != end && !is_space( c = *cur) )
      {
        word += c;
        ++cur;
      }

      l->push_back(word);

      ++count;

      --at_most;
    }

    return count;
  }



  template<class IsSeparator>
  static
  int  parse_words(string const       &s,
                   stringlist_t       *l,
                   unsigned            at_most,
                   IsSeparator const & is_space
                  )
    /// Parse a string, which is assumed to be collection of words
    /// separated by spaces.  The parsed words are stored in a list
    /// and the count of words is returned.  Note that leading and trailing
    /// spaces are removed.  This is a parse operation, not a split operation.
    /// If you want a split, rather than a parse, pick a different function.
    /// In this function, one or more spaces are considered to be "one space".
    ///
    /// You can specify a maximum limit of words to store in the list -- the 
    /// default is to store them all -- well 2 billion anyway.  If the maximum 
    /// limit is reached, all unparsed text is left in the last item in the
    /// list -- leading spaces will be removed from it, however.  That is,
    /// this function works much like the bourne shell 'read' command that
    /// reads words from a a file and stores them in environment variables.
    /// In this case, the parsed words are stored in a list of strings.
    ///
    /// The parameters are as follows:
    ///
    ///   @param s the string to be parsed
    ///
    ///   @param l a pointer to the list in which the parsed words will go
    ///            \note  words are appended to the end of the list, you will
    ///            have to clean it out of anything you don't want to appear
    ///            at the beginning of the list.
    ///
    ///   @param at_most the maximum number of elements to be parsed into
    ///                  the list, see note above about how unparsed text
    ///                  is handled.
    ///
    ///   @param is_space a functor describing how to tell if a character is
    ///                   a 'space' or not.  To make your own definitions, use
    ///                   StrTool::Is_Space as an example.
  {
    string::const_iterator cur = s.begin();
    string::const_iterator end = s.end();

    return parse_words(cur, end, l, at_most, is_space);

  }


                        
  static int  parse_words(string const  &s,  stringlist_t  *l )
    ///
    /// An overload of the parse_words function that eliminates seldom
    /// used parameters in favor of common defaults.  See the commentary
    /// above for the use of this function
  {
    return parse_words(s, l, all, Is_Space() );
  }

  static int  parse_words(string const  &s,  stringlist_t  *l, unsigned at_most )
    ///
    /// An overload of the parse_words function that eliminates seldom
    /// used parameters in favor of common defaults.  See the commentary
    /// above for the use of this function
  {
    return parse_words(s, l, at_most, Is_Space() );
  }

  static int  parse_lines(string const  &s,  stringlist_t  *l )
    ///  Parse a string into a list of lines separated by newline characters.
  {
    return parse_words(s, l, all, Is_Newline() );
  }

  static int  parse_lines(string const  &s,  stringlist_t  *l, unsigned at_most )
    //
    ///  Parses a string into a list of lines -- up to a count of at most
    ///  'at_most'.  The last parm will have all remaining text
    //
  {
    return parse_words(s, l, at_most, Is_Newline() );
  }


  static string expand_tabs(char const *sourceBytes,
                            bool* non_printable_chars = 0,
                            char  replacement = 0
                           );
    ///<
    /// Copy the input string and expand tab characters in the put
    /// into spaces in the output.  Tab expansion is done based on
    /// tab 'stop' positions -- it is not a simple string substitution.
    /// Basically, a tab character means 'insert spaces until the
    /// the text tab stop'.  Tab stops are 8 characters apart in the
    /// unix standard -- which the function implements.
    ///
    /// Parameters:
    ///
    /// @param sourceBytes           defines the text to be tab expanded.
    ///
    /// @param non_printable_chars   if specified must point to a boolean
    ///                              variable which will be modified to
    ///                              a value which indicates that non-printable
    ///                              characters exist in the sourceBytes.
    ///
    /// @param replacement           if specified, indicates that non-printable
    ///                              characters should be converted to 
    ///                              the value of replacement.


  static string expand_tabs(string const &r,
                            bool *non_printable_chars=0,
                            char replacement=0
                           )
    ///
    /// Same as the above, except it takes a std::string instead of a
    /// character buffer.
  {
    return expand_tabs(r.c_str(), non_printable_chars, replacement);
  }

  static string pack_tabs(char const *s, bool stop_after_first_non_blank=true);
    ///
    /// Convert spaces into tabs wherever possible according to the
    /// standard unix 8 character rule.  This is the reverse of 'expand_tabs'
    /// above.
    ///
    /// If stop_after_first_non_blank is true, then tab compression will stop
    /// as soon as the first non-blank character is encountered.  This prevents
    /// the accidental insertion of tab characters into quoted strings in
    /// program source code as can occur if you set don't set the flag.

  static string pack_tabs(string const &r, bool stop_after_first_non_blank=true)
    ///
    /// Same as the above, except it takes a std::string instead of a
    /// character buffer.
    ///
    /// If stop_after_first_non_blank is true, then tab compression will stop
    /// as soon as the first non-blank character is encountered.  This prevents
    /// the accidental insertion of tab characters into quoted strings in
    /// program source code as can occur if you set don't set the flag.
  {
    return pack_tabs(r.c_str());
  }

  template<class Char, class Iterator>
  static
  bool
  is_word_char(Char c, Iterator word_start, Iterator word_end)
    ///
    ///  Return true if a character is found in a sequence specified
    ///  by a pair of iterators.  The iterators are presumed to define
    ///  a list of valid word characters.
  {
    return std::find(word_start, word_end, c) != word_end;
  }


  template<class It1, class It2>
  static
  It1
  find_next_word(It1 start, It1 end, It2 word_start, It2  word_end)
    ///
    /// Search the sequence from start to end looking for the beginning
    /// of the next "word", from the given start position.  The sequence
    /// word_start to word_end defines a list of valid word characters.
    ///
    /// This function returns the beginning of the next word or it returns
    /// 'end'.  Assuming that these iterators are referring to characters,
    /// here is an example of word start positions:
    ///
    /// @code
    ///   word1   word2 word3                     word4
    ///   ^       ^     ^                         ^
    /// @endcode
    /// Only positions noted by carets will be returne.d
  {

    // skip past word characters under the cursor

    while(start != end)
    {
       if(!is_word_char(*start, word_start, word_end))
       {
         break;
       }

       ++start;

    }

    // scan to the beginning of the next word

    return std::find_first_of(start, end, word_start, word_end);

  }


  template<class It1, class It2>
  static
  It1
  find_prev_word(It1 start, It1 begin, It1 end, It2 word_start, It2 word_end)
    ///
    /// Move backwards from start to the beginning of the current or previous
    /// word.  Words are defined by the characters in the range word_start
    /// to word_end.  See find_next_word above for a description of how
    /// word start positions appear.
  {

    if(start == begin)
      return start;

    if(start == end)
      --start;

    if(start == begin)
      return start;

    if( is_word_char(*start, word_start, word_end) )
    {
      // find beginning of current word -- unless we are
      // already at the beginning, then find the beginning of the
      // previous word

      --start;

      if(start == begin)
        return start;

      if(is_word_char(*start, word_start, word_end))
      {
        // find beginning of this word

        --start;

        while(start != begin)
        {
          if( !is_word_char(*start, word_start, word_end) )
          {
            ++start;
            break;
          }
        
          --start;
        }

        // we are either at the beginning of the word or the
        // beginning of the string

        return start;

      }
      else
      {
        // drop through and find beginning of previous
        // word.
      }

    }

    // we are not in a word character -- scan back to the
    // beginning of the previous word.

    while(start != begin)
    {
      if(is_word_char(*start, word_start, word_end))
      {
        break;
      }
      --start;
    }

    if(start == begin)
      return start;

    while(start != begin)
    {
      if(!is_word_char(*start, word_start, word_end))
      {
        ++start;
        return start;
      }
      --start;
    }

    return start;

  }

  static void remove_leading(std::string &s, char space=' ');
    ///<  Remove all leading copies of space from the specified string.

  static void trim(std::string &s, char const *chars=" \t\r\n");
     ///<  Remove leading and trailing spaces, or other 

  static int compareInsensitive(std::string const &a, std::string const &b)
     ///<  Compare a and b ignoring case.  
     ///   
     ///   Note that if you are using this function as a parameter to the
     ///   as a sorter for a container, you must remember to handle
     ///   the possibility of having multiple strings in the container with
     ///   with different cases.  If you ignore that possibility then you
     ///   only be able to insert one object of a given "insensitive value"
     ///   in your container.  But that is  a problem for the sort routine
     ///   and should not in this function.
  {
     size_t asize = a.size();
     size_t bsize = b.size();

     size_t smallerSize = asize < bsize ? asize : bsize;

     int rc = STRNCASECMP(a.data(), b.data(), smallerSize);

     if(rc != 0)
     {
        return rc;
     }

     // if you get here, a and b are equal for some leading number of characters
     // so define the shorter of the two total strins as being less

     if(asize < bsize)
        return -1;
     else
     if(asize > bsize)
        return 1;

     // a and b are the same length and they are of equal when we get here



     return 0;

     
  }


   static bool snatch_file_info(std::string const &s, std::string &file, int &line);
   ///<
   ///  Snatch file information out of a string and return it in @b file and
   ///  @b line.
   ///
   ///  The input string is assumed to be one of the following formats:
   ///
   ///  @li  @b filename:lineNo:      -- grep format
   ///  @li  @b file @b lineNo:       -- borland c output
   ///  @li  @b "file", @b line @b 2: -- other compiler outputs
   ///  @li  @b file(line)            -- microsoft
   ///  @li  @b \#line @b 12 @b "file" -- C preprocessor (some compilers)
   ///  @li  @b &lt;file&gt;          -- \#include file names
   ///  @li  @b #12 @b "file"         -- C preprecessor (other compilers)
   ///  @li  @b Error @b 112: @b "reports_pkg.c", @b line @b 48  
   ///  @li  @b Error @b (112): @b "reports_pkg.c", @b line @b 48 
   ///  @li  @b error: @b "reports_pkg.c", @b line @b 48
   ///  @li  @b warning: @b "reports_pkg.c", @b line @b 48
   ///  @li  @b file:  -- grep without the -n
   ///
   ///  @param[in]  s 
   ///              The string containing the source information
   ///
   ///  @param[out] file
   ///              The output file name, if any.
   ///
   ///  @param[out] line
   ///              The output line number if any
   ///
   ///  @return True if successful, but if line is 0, then the line number should
   ///               be ignored.


  struct CharCompare
      ///  Functor object used to compare characters either case sensitively or
      ///  or insentively.
      ///
      ///  Use this class like this:
      ///
      ///    CharCompare   comparator(true);  // case insensitive comparisons
      ///
      ///    bool  equal = comparator('a', 'A');  
      ///
      ///    char const *first="abc";
      ///    char const *second="BC";
      ///
      ///    char const *bc = std::search(first, first+3, second, second+3, comparator);
      ///
      ///  The pointer, bc, now points to the "bc" in first.
  {
    bool insensitive_;

    bool operator() (char a, char b) const
    {
        if(insensitive_)
        {
          if( isupper(a) )
            a = tolower(a);
        
          if( isupper(b) )
            b = tolower(b);
        
          return a == b;
        }
        
        return a == b;
    }

    CharCompare(bool caseInsensitive)
    : insensitive_(caseInsensitive)
    {
    }

  };

  static std::string lowercase(std::string const &rhs)
  {
     std::string rv;

     rv.reserve(rhs.size());

     char const *f = rhs.data();
     char const *l = f + rhs.size();

     while(f != l)
     {
        char a = *f++;

          if( isupper(a) )
            a = tolower(a);

        rv += a;
     }


     return rv;
  }

  static std::string lowercase(char const *rhs)
  {
     std::string rv;

     size_t length = strlen(rhs);

     rv.reserve(length);

     char const *f = rhs;
     char const *l = f + length;

     while(f != l)
     {
        char a = *f++;

          if( isupper(a) )
            a = tolower(a);

        rv += a;
     }


     return rv;
  }


  static std::string uppercase(std::string const &rhs)
  {
     std::string rv;

     rv.reserve(rhs.size());

     char const *f = rhs.data();
     char const *l = f + rhs.size();

     while(f != l)
     {
        char a = *f++;

          if( islower(a) )
            a = toupper(a);

        rv += a;
     }


     return rv;
  }

  static std::string uppercase(char const *rhs)
  {
     std::string rv;

     size_t length = strlen(rhs);

     rv.reserve(length);

     char const *f = rhs;
     char const *l = f + length;

     while(f != l)
     {
        char a = *f++;

          if( islower(a) )
            a = toupper(a);

        rv += a;
     }


     return rv;
  }




  static size_t find(char const *& fragment, size_t fragLen, char const *&within, size_t totalLen, bool insensitive)
      ///
      ///    search for a string fragment within a larger string and return an offset to the start of the
      ///    fragment.  If not found, then return an offset outside the larger string.
      ///
      ///    @param[in]  fragment The begining of the substring to be searched for.
      ///    @param[in]  fragLen  The length of the substring to be searched for
      ///    @param[in]  within   The larger string from which to perform the search
      ///    @param[in]  totalLen The length of the larger string from the search is being performed
      ///    @param[in]  insensitive  True implies a case insensitive search.
      ///
      ///    @return  The offset within the larger string to the point where the fragment is first found.
      ///             If not found, the offset will be within+totalLen.
      ///
  {
      
      char const *where = std::search( within, within+totalLen, fragment, fragment+fragLen, CharCompare(insensitive));

      return where - within;

  }

  static size_t find(std::string const &fragment, std::string const &within, bool insensitive)
  {
     char const *frag = fragment.data();
     char const *with = within.data();

     return find(frag, fragment.size(), with, within.length(), insensitive);
  }

  static size_t find(char const *fragment, std::string const &within, bool insensitive)
  {
     char const *with = within.data();
     return find(fragment, strlen(fragment), with, within.length(), insensitive);
  }


  template<class Iterator, class Iterator2>
  static inline void skipBlanks(Iterator &first, Iterator2 &last)
      /// move an iterator past leading whitespace characters.
      ///
      /// @param[in,out] first   The iterator to be moved.
      /// @param[in]     last    The iterator that defines the end of the range.
      ///
      /// @note  The two iterators are passed in as references and do not have to
      ///        be the same time.  Notably, the first can be an "iterator" member 
      ///        of an STL container and the second iterator can be a "const_iterator".
  {
      while(first != last)
      {
         char c = *first;

         if(!isspace(c))
            break;

         ++first;
      }
  }


  static
  bool getline(FILE* file, std::string &line);   ///<  Read a line of unspecified size.
                                                 ///   @param file  The file pointer returned by fopen, popen, etc
                                                 ///   @param line  The location to store the line.
                                                 ///
                                                 ///   @return true if successful.  False if error or end of file.



  static
  void fold(std::string const &input, size_t maxLen, std::vector<std::string> &output);  ///<  Split the specifed input line into
                                                                                         ///   multiple lines if it bigger than maxLen
                                                                                         ///   based on the availability of non-word
                                                                                         ///   characters to split upon.

};  // class StrTool

} // namespace cxxtls

#endif
