//
// Copyright 2014, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include <stdint.h>
#include <memory>
#include <string>
#include <string.h>

namespace cxxtls
{
namespace utf
{
    ///  A few functions designed to make it easier to live in a unicode world.
    ///  this is not a replacement for a real unicode handling toolkit.


    unsigned char conflate(wchar_t);  ///<  @return a normal ANSI character that is a rough approximation
                                      ///    to the utf characters from the european language variants.
                                      ///    basically, just return trash for everything else.
                                      ///
                                      ///   @param[in]  utfCharCode

    enum encoding
    {
        // future maintainers, these numbers must run from 0 to ENCODING_TYPE_COUNT and there
        // must be no gaps or numbers out of order
        //
        // Also:
        //
        //  Keep LineDecoder updated as you modify this list.  See below.
        //
        //
        utf8_enc,
        utf16be_enc,  // 16 bit, big endian
        utf16le_enc,
        utf32be_enc,
        utf32le_enc,
        utfunk_enc,  // unknown -- encoding is not clear
        utf7_enc,    // ansi encoding
        utfbin_enc,  // 8 bit data found that is not valid unicode
        ENCODING_TYPE_COUNT  // must be last
    };

   char const *encodingName(encoding);  ///<  @return the name of the encoding type.  Names are as follows 
                                        ///    utf8, utf16be, utf16le, utf32be, utf32le, utfunk.
       

    typedef uint32_t   utfchar_t;  ///<  UTF representation of a character.  21 bits worth of data.  
                                   ///   Note that the UTF code matches the ANSI characters for the range
                                   ///   blank - tilde.
                                   ///
                                   ///   Note also that wchar_t may not be the same as this.  It might only be
                                   ///   16 bits -- such as on windows.


//--------------------------------------------------------------------------------------------------------------------
// From http://www.unicode.org/faq/utf_bom.html
//
//  Name                            UTF-8   UTF-16  UTF-16BE        UTF-16LE        UTF-32  UTF-32BE        UTF-32LE
//  Smallest code point             0000    0000    0000            0000    0000    0000    0000
//  Largest code point              10FFFF  10FFFF  10FFFF          10FFFF  10FFFF  10FFFF  10FFFF
//  Code unit size  8 bits          16 bits         16 bits         16 bits         32 bits 32 bits         32 bits
//  Byte order                      N/A     <BOM>   big-endian      little-endian   <BOM>   big-endian      little-endian
//  Fewest bytes per character      1       2       2               2               4       4               4
//  Most bytes per character        4       4       4               4               4       4               4
//--------------------------------------------------------------------------------------------------------------------


//--------------------------------------------------------------------------------------------------------------------
// From http://www.unicode.org/faq/utf_bom.html
//
// Bytes           Encoding Form
// 00 00 FE FF     UTF-32, big-endian
// FF FE 00 00     UTF-32, little-endian
// FE FF           UTF-16, big-endian
// FF FE           UTF-16, little-endian
// EF BB BF        UTF-8
//--------------------------------------------------------------------------------------------------------------------

   
   template<class Iterator>
   encoding   streamEncoding(Iterator &first, Iterator last)
   {
       ///
       ///  @param[in,out]  first -- a pointer to the first byte to be examined.
       ///  @param[in]      last  -- a pointer past the last byte to be examined.
       ///
       ///  @return the encoding type of the stream and leave the iterator, first, pointing to the first
       ///          byte after the byte order mark.

       Iterator save = first;

       if(first == last)
          return utf7_enc;  // can't know -- call it ANSI text

       unsigned char c = (*first++);

       switch(c)
       {
           default:
           {
               // first character is not a valid byte order mark character

               first = save;

               while(save != last)
               {
                   c = *save++;

                   if(c >= ' ' && c <= '~')  // ansi characters
                      continue;

                   if(c > '~')
                      return utfbin_enc;

                   if(! (
                               c == 8      // backspace
                            || c == 9      // tab
                            || c == '\r'   // return
                            || c == '\n'   // newline
                        )
                     )
                   {
                       return utfbin_enc;
                   }

               }

               
               return utf7_enc; // plain old ascii
           }

           case 0x00:
           {
               // possible utf32be_enc

               if(first == last)
               {
                  first = save;

                  return utfunk_enc;
               }

               c = (*first++);

               if(c != 0x00)
               {
                  first = save;
                  return utfunk_enc;
               }


               if(first == last)
               {
                  first = save;
                  return utfunk_enc;
               }

               c = (*first++);

               if( c != 0xFE)
               {
                  first = save;

                  return utfunk_enc;
               }

               if(first == last)
               {
                  first = save;
                  return utfunk_enc;
               }
               
               c = (*first++);

               if(c != 0xFF)
               {
                   first = save;

                   return utfunk_enc;

               }

               return utf16be_enc;
           }
           break;

           case 0xff:
           {
               // possible le 16 or le 32

               if(first == last)
               {
                  first = save;

                  return utfunk_enc;
               }

               c = (*first++);

               if( c != 0xFE )
               {
                  first = save;
                  return utfunk_enc;
               }

               if(first == last)
               {
                  return utf16le_enc;  // data only contains an encoding type with no data
               }

               save = first; 

               c = (*first++);

               if(c == 0)
               {
                  // possibly have a 32 bit encoding.

                  if(first == last)
                  {
                     first = save;
                     return utf16le_enc;
                  }

                  c = (*first++);

                  if(c == 0x00)
                  {
                     return utf32le_enc;
                  }

                  // drop through to execute the first = save and return of 16 be.

               }

               first = save;

               return utf16le_enc; // nope, plain old 16 bit little endian

           };
           break;
           
           case 0xfe:
           {
               if(first == last)
               {
                  first = save;
                  return utfunk_enc;
               }

               c = (*first++);

               if(c == 0xff)
               {
                  return utf16be_enc;
               }

               first = save;

               return utfunk_enc;

           }
           break;

           case 0xef:
           {
              if(first == last)
              {
                  first = save;
                  return utfunk_enc;
              }

              c = (*first++);

              if(c != 0xbb)
              {
                 first = save;

                 return utfunk_enc;
              }

              if(first == last)
              {
                 first = save;

                 return utfunk_enc;
              }

              c = (*first++);

              if(c != 0xbf)
              {
                 first = save;

                 return utfunk_enc;
              }

              return utf8_enc;

           }
           break;
       }

   }


   inline encoding stringEncoding(char const *&p)
      /// Get the encoding type from p, moving p past the byte order mark
   {
       size_t l = strlen(p);

       return streamEncoding(p, p+l);
       
   }

   encoding fileEncoding(char const *name);
       /// Get the encoding by reading the first few bytes of a file.


   struct decoder_base
   {
      virtual ~decoder_base() {}      // do nothing base class for decoders to help clean up memory
   };

   template<encoding EncType>
   struct decoder
   :  decoder_base
      // unimplemented decoder.   This is what you get when you use an unimplemented decoder
      // 
   {
       enum constants { implemented_ = 0 };


       utfchar_t operator() (char const *&p, char const *end)
       {
          if(p == end)
             return -1;

          return *p++;
       }

       utfchar_t operator() (unsigned char const *&p, unsigned char const *end)
       {
          if(p == end)
             return -1;

           return *p++;
       }

       virtual ~decoder() {}
   };


   template<>
   struct decoder<utf7_enc>
   :  decoder_base
       ///  A class object that acts like a function.  This functor has the job of 
       ///  parsing a single utf character out of a byte stream and returning 
       ///  either -1 if the encoding is bad or returning the proper utf character
       ///  code.  If the function is successfull, it will increment the pointer
       ///  passed into it.  If not, it will not.
       ///
       ///  There will be one of these encoders for each encoding defined by the
       ///  utf committee.
       ///
       ///  This particular class only handles ANSI characters being decoded as
       ///  UTF, so the decoder just returns the character straight back.
   {
       enum constants { implemented_ = 1 };    // use this to see if the decoder type is implemented

       decoder()
       {
       }

       utfchar_t operator() (char const *&p, char const *end)
       {
          if(p == end)
             return -1;
          return *p++;
       }

       utfchar_t operator() (unsigned char const *&p, unsigned char const *end)
       {
          if(p == end)
             return -1;
          return *p++;
       }

       virtual ~decoder() {}

   };

   template<>
   struct decoder<utf8_enc>
   :  decoder_base
       ///  A class object that acts like a function.  This functor has the job of 
       ///  parsing a single utf character out of a byte stream and returning 
       ///  either -1 if the encoding is bad or returning the proper utf character
       ///  code.  If the function is successfull, it will increment the pointer
       ///  passed into it.  If not, it will not.
       ///
       ///  There will be one of these encoders for each encoding defined by the
       ///  utf committee.
       ///
       ///  This particular class only handles the utf8 encoding.
   {
       // design borrowed from Nemanja Trifunovic but this code does not work
       // like his.

       enum constants { implemented_ = 1 };    // use this to see if the decoder type is implemented


       decoder()
       {
       }

       utfchar_t operator() (char const *&p, char const *end)
       {
          unsigned char const *&r = reinterpret_cast<unsigned char const *&>(p);
          unsigned char const *s  = reinterpret_cast<unsigned char const *>(end);

          return this->operator()(r,s);
       }

       utfchar_t operator() (unsigned char const *&p, unsigned char const *end)
       {
           

           unsigned char const *save = p;

           utfchar_t head = *p;               // utfchar_t is unsigned

           if (head < 0x80 )
               return *p++;              // normal ansi characters
           else 
           if ((head >> 5) == 0x6)
           {
              // two character pattern

              ++p;

              if(p == end)
              {
                 p = save;
                 return utfchar_t(-1);
              }

              return ((head << 6) & 0x7ff) + ((*p++) & 0x3f);

           }
           else 
           if ((head >> 4) == 0xe)
           {
              // 3 character pattern

              utfchar_t cp = 0;

              cp = *p++ & 0xff;

              if(p == end)
              {
                 p = save;
                 return utfchar_t(-1);
              }
        
              cp = ((cp << 12) & 0xffff) + (((*p++ & 0xff) << 6) & 0xfff);

              if(p == end)
              {
                 p = save;
                 return utfchar_t(-1);
              }

              cp += (*p++) & 0x3f;

             return cp;
           }
           else 
           if ((head >> 3) == 0x1e)
           {
              // 4 character pattern


              utfchar_t cp = *p++ & 0xff;

              if(p == end)
              {
                 p = save;
                 return utfchar_t(-1);
              }


              cp = ((cp << 18) & 0x1fffff) + (((*p++ & 0xff) << 12) & 0x3ffff);

              if(p == end)
              {
                 p = save;
                 return utfchar_t(-1);
              }


              cp += ((*p++ & 0xff) << 6) & 0xfff;

              if(p == end)
              {
                 p = save;
                 return utfchar_t(-1);
              }


              cp += (*p++) & 0x3f;

              return cp;
           }
           else
               return -1;    // invalid character encoding.

       }

       virtual ~decoder() {}


   }; // struct decoder<utf8_enc>

   template<>
   struct decoder<utf16be_enc>
   :  decoder_base
       ///  A class object that acts like a function.  This functor has the job of 
       ///  parsing a single utf character out of a byte stream and returning 
       ///  either -1 if the encoding is bad or returning the proper utf character
       ///  code.  If the function is successfull, it will increment the pointer
       ///  passed into it.  If not, it will not.
       ///
       ///  There will be one of these encoders for each encoding defined by the
       ///  utf committee.
       ///
       ///  This particular class only handles the utf16be encoding.
   {
       enum constants { implemented_ = 1 };    // use this to see if the decoder type is implemented

       decoder()
       {
           pushed_back_code_unit_ = 0;
       }

       utfchar_t operator() (char const *&p, char const *end)
       {
          unsigned char const *&r = reinterpret_cast<unsigned char const *&>(p);
          unsigned char const *s  = reinterpret_cast<unsigned char const *>(end);

          return this->operator()(r,s);
       }

       unsigned pushed_back_code_unit_;

       
       unsigned read_unsigned_short(unsigned char const * &p)
       {
          /// either grap the pushed back code unit or parse a 2 byte unsigned word out of the input
          /// stream.  Don't check for out of data, the caller must do that.

          if(pushed_back_code_unit_)
          {
             unsigned tmp = pushed_back_code_unit_;

             pushed_back_code_unit_ = 0;

             return tmp;
          }

          unsigned char tmp = *p++;
          
          return (tmp << 8)|*p++;


       }

       bool outOfData(unsigned char const *p, unsigned char const *end)
       {
           if(   !pushed_back_code_unit_ 
              && ((end - p) < 2)
             )
              return true;

           return false;
       }


       void push_back(unsigned cu)
       {
          pushed_back_code_unit_ = cu;
       }


       utfchar_t operator() (unsigned char const *&p, unsigned char const *end)
       {
          // from wikipedia

              if(outOfData(p,end))
                  return -1;

              unsigned code_unit = read_unsigned_short(p);

              if (code_unit >= 0xD800 && code_unit <= 0xDBFF) 
              {
                if(outOfData(p,end))
                {
                   return -1;
                }

                unsigned code_unit_2 = read_unsigned_short(p);

                if (code_unit_2 >= 0xDC00 && code_unit_2 <= 0xDFFF)
                  return (code_unit << 10) + code_unit_2 - 0x35FDC00;

                push_back(code_unit_2);
              }

              return code_unit;

       }

       virtual ~decoder() {}

   }; // decoder<utf16be_enc>
   

   template<>
   struct decoder<utf16le_enc>
   :  decoder_base
       ///  A class object that acts like a function.  This functor has the job of 
       ///  parsing a single utf character out of a byte stream and returning 
       ///  either -1 if the encoding is bad or returning the proper utf character
       ///  code.  If the function is successfull, it will increment the pointer
       ///  passed into it.  If not, it will not.
       ///
       ///  There will be one of these encoders for each encoding defined by the
       ///  utf committee.
       ///
       ///  This particular class only handles the utf16le encoding.
   {
       enum constants { implemented_ = 1 };    // use this to see if the decoder type is implemented

       decoder()
       {
           pushed_back_code_unit_ = 0;
       }

       utfchar_t operator() (char const *&p, char const *end)
       {
          unsigned char const *&r = reinterpret_cast<unsigned char const *&>(p);
          unsigned char const *s  = reinterpret_cast<unsigned char const *>(end);

          return this->operator()(r,s);
       }

       unsigned pushed_back_code_unit_;

       
       unsigned read_unsigned_short(unsigned char const * &p)
       {
          /// either grap the pushed back code unit or parse a 2 byte unsigned word out of the input
          /// stream.  Don't check for out of data, the caller must do that.

          if(pushed_back_code_unit_)
          {
             unsigned tmp = pushed_back_code_unit_;

             pushed_back_code_unit_ = 0;

             return tmp;
          }

          unsigned char tmp = *p++;
          
          return (*p++ << 8) | tmp;


       }

       bool outOfData(unsigned char const *p, unsigned char const *end)
       {
           if(   !pushed_back_code_unit_ 
              && ((end - p) < 2)
             )
              return true;

           return false;
       }


       void push_back(unsigned cu)
       {
          pushed_back_code_unit_ = cu;
       }


       utfchar_t operator() (unsigned char const *&p, unsigned char const *end)
       {
          // from wikipedia

              if(outOfData(p,end))
                  return -1;

              unsigned code_unit = read_unsigned_short(p);

              if (code_unit >= 0xD800 && code_unit <= 0xDBFF) 
              {
                if(outOfData(p,end))
                {
                   return -1;
                }

                unsigned code_unit_2 = read_unsigned_short(p);

                if (code_unit_2 >= 0xDC00 && code_unit_2 <= 0xDFFF)
                  return (code_unit << 10) + code_unit_2 - 0x35FDC00;

                push_back(code_unit_2);
              }

              return code_unit;

       }

       virtual ~decoder() {}

   }; // decoder<utf16be_enc> 


   template<encoding Type>
   struct LineDecoderImpl
      /// implementation of the line decoder algorithm
   {
       decoder<Type> dec_;

       template<class Iterator1, class Iterator2>
       std::wstring decode(Iterator1 &scan, Iterator2 const &end)
       {
           std::wstring value;

           while(scan < end)
           {
               unsigned codePoint = dec_(scan, end);
               
               if(codePoint == '\n')
                  break;

               value += wchar_t(codePoint);
           }

           return value;
       }

   };

   struct LineDecoder
   {
       LineDecoderImpl<utf8_enc>     utf8_enc_impl;
       LineDecoderImpl<utf16be_enc>  utf16be_enc_impl;
       LineDecoderImpl<utf16le_enc>  utf16le_enc_impl;
       LineDecoderImpl<utf32be_enc>  utf32be_enc_impl;
       LineDecoderImpl<utf32le_enc>  utf32le_enc_impl;
       LineDecoderImpl<utf7_enc>     utf7_enc_impl;
       LineDecoderImpl<utfunk_enc>   utfunk_enc_impl;
       LineDecoderImpl<utfbin_enc>   utfbin_enc_impl;

       template<class Iterator1 , class Iterator2>
       std::wstring decode(encoding type, Iterator1 &scan, Iterator2 const &end)
       {
           switch(type)
           {
                case utf8_enc:       return utf8_enc_impl.decode(scan,end);
                case utf16be_enc:    return utf16be_enc_impl.decode(scan,end);
                case utf16le_enc:    return utf16le_enc_impl.decode(scan,end);
                case utf32be_enc:    return utf32be_enc_impl.decode(scan,end);
                case utf32le_enc:    return utf32le_enc_impl.decode(scan,end);
                case utfunk_enc:     return utfunk_enc_impl.decode(scan,end);
                case utf7_enc:       return utf7_enc_impl.decode(scan,end);
                default:             return utfbin_enc_impl.decode(scan,end);
           }
       }

   };




};  //  namespace utf
};  //  namespace cxxtls

