//
// Copyright 2013, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
#ifdef __linux__
#include <cxxtls/dumbterminal.h>
#include <portable_io.h>
#include <cxxtls/viewermanager.h>
#include <cxxtls/terminalstatus.h>
#include <cxxtls/strtool.h>
#include <cxxtls/foreach.h>
#include <fstream>
#include <unistd.h>
#include <errno.h>
#include <pty.h>
#include <sys/ioctl.h>


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
namespace cxxtls
{
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
DumbTerminalViewer::
~DumbTerminalViewer()
{
    {
       ScopedMutexLock lock(shutDownMutex_);
       shuttingDown_ = true;
    }


    // before joining the thread, we need to do something to terminate
    // the process that might be reading/writing the slave device, because
    // if we don't, the reader function, below, will not get an end of
    // file and thus the joinThread call below will hang.  A short hang is
    // not bad, but a long hang is.
    //
    // Note that a long hang will occur if you have a process reading 
    // from the slave terminal and it does not terminate.  You need to 
    // orchestrate the termination of that process outside the scope of the
    // dumb terminal viewer.
    //
    // I tried various things that I thought would work but they didn't --
    // at least not completely.
    //
    // I left one attempt in the code because it does work on bash -- that
    // is, if you write a ^D character to the masterFD, bash, if reading
    // from the slave device will interpret it as end of command input and
    // exit -- thus allowing the join, below, to terminate quickly.  However,
    // this does not work if you care running 'cat </dev/pts/1'.  The cat
    // program does 

    {
        static char ctrlD[2]={ 0x04, '\n' };
        
        write(masterFD_, &ctrlD[0], 2);
        
        // close(masterFD_); // signal read thread to terminal

        close(masterFD_);  // will terminate the reader() thread

    }

    joinThread(thread_);  // join the now terminating thread thread


}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
    
void 
DumbTerminalViewer::
operator() ( CursorWindow::viewport *     vp,
             int                          cmd
           )
{
    if(suppressRepaint_)
    {
       cmdType_ = cmd;
       paintTitle(vp);
       return;
    }


    ScopedMutexLock lock(outputMutex_);

    getWindowSize(vp, cmd);

    {
       struct winsize ws;

          ws.ws_row = windowRows_;
          ws.ws_col = windowCols_;
          ws.ws_xpixel = windowCols_;
          ws.ws_ypixel = windowRows_;

       ioctl(masterFD_, TIOCSWINSZ, &ws);
       
    }

    paintWindow(vp);


}  // operator() ()

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

Viewer::BackgroundProcessStatus 
DumbTerminalViewer::
getBackgroundData(CursorWindow::input_event &e)
{
    bool dataAvailable;

    {
       ScopedMutexLock lock(dataArrivedMutex_);

       dataAvailable = dataArrived_;

       dataArrived_=false;

    }

    {
       ScopedMutexLock lock(shutDownMutex_);

       if(shuttingDown_)
       {
          dataAvailable = true;   // treat shutdown request as data available,
                                  // handle_event checks for shuttingDown_ and
                                  // returns true when it arrives
       }
    }

    if(dataAvailable)
    {
        e.type_ = CursorWindow::input_event::BackgroundData;

        return BACKGROUND_DATA_AVAILABLE;
    }


    return  NO_BACKGROUND_DATA_AVAILABLE;

}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

bool 
DumbTerminalViewer::
handle_event( CursorWindow::input_event const * e,
              CursorWindow::viewport          * vp
            )
{
    ScopedMutexLock lock(outputMutex_);

    {
        ScopedMutexLock lock(shutDownMutex_);

        if(shuttingDown_)
           return true;
    }
    getWindowSize(vp, cmdType_);

    if(   (e->type_ != CursorWindow::input_event::ForceExitKey) 
       && (e->type_ != CursorWindow::input_event::BackgroundData)
       && ctrlX_
      )
    {
       // We are in the ctrlX_ true state -- meaning that the user
       // has just pressed ctrlX_ and we have not found another character that
       // should follow it naturally.
       //
       // Thus, we are in a "custom" function key sequence.

       ctrlX_ = false;

       switch(e->value_)
       {
          case 0x0b:   // ^X ^K
              return true; 

          case 0x15:   // ^X ^U
              if(inputRows_ < windowRows_ - titleRows_ - 1)
                  ++inputRows_;
              getWindowSize(vp, cmdType_);
              paintWindow(vp);
              break;

          case 0x19:   // ^X ^Y
              {
                  if(inputRows_ > 1)
                     --inputRows_;
                  
                  int neededRows = inputArea();
                  
                  if(inputRows_ < neededRows)
                     inputRows_ = neededRows;
                  
                  getWindowSize(vp, cmdType_);
                  paintWindow(vp);
              }
              break;


           default:
              break;
       }

       return false;
    }

    if(e->type_ == CursorWindow::input_event::ForceExitKey)
    {
       ctrlX_ = false;
       return true;
    }
    else
    if(e->type_ == CursorWindow::input_event::BackgroundData)
    {
        // background data does not come through the event mechanism -- only the fact
        // that background data exists matters.  The output_ page is updated
        // in the background by the reader() (a separate thread) and the only
        // thing we need to do here is to paint the screen to acknowledge its
        // existence

        paintWindow(vp);
    }
    else
    if(e->type_ == CursorWindow::input_event::FunctionKey)
    {
        int edit_func = vp->window()->func[e->value_];

        switch(edit_func)
        {
          case CursorWindow::func_top:
                    if(!output_.empty())
                    {
                      cursorRow_ = output_.size() - 1;
                    
                      if(size_t(scrollRows_) >= output_.size())
                      {
                          pageBottom_ = 0;
                      }
                      else
                      {
                         pageBottom_ = cursorRow_ - (scrollRows_ - 1);
                      }
                    
                      getWindowSize(vp, cmdType_); 
                      paintWindow(vp);
                    }
              break;
        
          case CursorWindow::func_bottom:
                  if(!output_.empty())
                  {
                      pageBottom_ = 0;
                      cursorRow_  = 0;
                      getWindowSize(vp, cmdType_); 
                      paintWindow(vp);
                  }
              break;

          case CursorWindow::func_up:
            if(cursorRow_ == size_t(-1))
              {
                 // we are currently sitting in the input region
        
                 if(inputRow_ == 0)
                 {
                    cursorRow_ = 0;
                 }
                 else
                 {
                    --inputRow_;  // move up on the screen
                    inputOffset_ -= availInputCols_;
                    inputCol_ = inputOffset_ % availInputCols_ + leftInputCol_;
                 }
        
              }
              else
              if(cursorRow_ == pageBottom_ + scrollRows_ -1)
              {
                 if(cursorRow_ < output_.size())
                 {
                    ++pageBottom_;
                    ++cursorRow_;
                 }
              }
              else
              {
                 if(cursorRow_ < output_.size())
                      ++cursorRow_;
              }
        
              getWindowSize(vp, cmdType_); // and reset screen text attributes
              paintWindow(vp);
              break;
        
          case CursorWindow::func_down:
              if(cursorRow_ == 0)
              {
                 cursorRow_ = -1;
                 inputRow_  = 0;
                 // we have just moved into the input region.
        
              }
              else
              if(cursorRow_ == size_t(-1))
              {
                  if(inputRow_ < inputRows_ - 1)
                  {
                     ++inputRow_;
                     inputOffset_ += availInputCols_;
                     inputCol_ =  (inputOffset_ % availInputCols_) + leftInputCol_;
                  }
              }
              else
              if((cursorRow_ == pageBottom_) && (pageBottom_ != 0))
              {
                 --pageBottom_;
                 cursorRow_ = pageBottom_;
              }
              else
              {
                  --cursorRow_;
              }
              getWindowSize(vp, cmdType_); // and reset screen text attributes
              paintWindow(vp);
              break;
        
          case CursorWindow::func_left:
              if(cursorRow_ == size_t(-1))
              {
                 // we are in the input area -- move the cursor left

                 if(inputOffset_)
                 {
                     --inputOffset_;

                     if(inputCol_ > leftInputCol_)
                     {
                        --inputCol_;
                     }
                     else
                     {
                        --inputRow_;
                        inputCol_ = windowCols_-1;
                     }

                     paintCommandLine(vp);

                 }

              }
              else
              {
                  // we are in the scrolling area -- slide the window left

                  if(leftOffset_)
                      --leftOffset_;
                  paintWindow(vp);
              }
              break;
        
          case CursorWindow::func_right:
              if(cursorRow_ == size_t(-1))
              {
                  // we are in the input area, move the cursor right

                  if(   ( !inputLine_.empty() )
                     && ( size_t(inputOffset_) < inputLine_.size() )
                    )
                  {
                     // we are not after the empty cell at the end of the line text
                     
                     ++inputOffset_;

                     ++inputCol_;

                     if(inputCol_ == windowCols_)
                     {
                       ++inputRow_;
                       inputCol_ = leftInputCol_;
                     }

                     paintCommandLine(vp);

                  }

              }
              else
              {
                  // we are in the scrolling region -- slide the window right

                  ++leftOffset_;
                  paintWindow(vp);
              }
              break;
        
          case CursorWindow::func_home:

              if(cursorRow_ == size_t(-1))
              {
                   inputRow_ = 0;
                   inputCol_ = leftInputCol_;
                   inputOffset_ = 0;
                   paintWindow(vp);
              }
              else
              {
                   leftOffset_ = 0;
                   paintWindow(vp);
              }
              break;
        
          case CursorWindow::func_end:
              if(cursorRow_ != size_t(-1))
              {
                   // we are in the scrolling region.

                   // move the left offset to 1 page less than the
                   // widest line currently visible
                   
                size_t width = 0;

                   size_t outputSize = output_.size();
                   
                   for(int i = 0; i < scrollRows_; ++i)
                   {
                      if(pageBottom_ + i < outputSize)
                      {
                          size_t l = output_[pageBottom_+i].size();

                          if(l > width)
                             width = l;
                      }
                   }

                   if(width == size_t(-1) || width < size_t(availInputCols_))
                       leftOffset_ = 0;
                   else
                   {
                       leftOffset_ = width - availInputCols_;
                   }
                   paintWindow(vp);
              }
              else
              {
                 // we are in the input editing region.

                 inputOffset_ = inputLine_.size();

                 inputRow_ = inputOffset_ / availInputCols_;
                 inputCol_ = inputOffset_ % availInputCols_ + leftInputCol_;
                 paintCommandLine(vp);

              }
              break;
        
          case CursorWindow::func_prior:  // page back
              if(   cursorRow_ < output_.size() 
                 || cursorRow_ == size_t(-1)
                )
              {
                 if(cursorRow_ == size_t(-1))
                 {
                     cursorRow_ = scrollRows_ -1;
                 }
                 else
                 {
                     pageBottom_ += scrollRows_ -1;
                    
                     if(pageBottom_ >= output_.size())
                     {
                        if(output_.size())
                           pageBottom_ = output_.size()-1;
                        else
                           pageBottom_ = 0;
                     }
                    
                     cursorRow_ = pageBottom_ + scrollRows_ -1;
                    
                 }
                 getWindowSize(vp, cmdType_); // and reset screen text attributes
                 paintWindow(vp);

              }

              break;
        
          case CursorWindow::func_next:   // page next
              if(cursorRow_ > size_t(scrollRows_))
              {
                  if(cursorRow_ == size_t(-1))
                  {
                     // we are in the input region

                     break;  // do nothing

                  }

                  if(pageBottom_ > size_t(scrollRows_-1))
                  {
                       pageBottom_ -= scrollRows_ -1;
                  }
                  else
                       pageBottom_ = 0;

                  cursorRow_ = pageBottom_ + scrollRows_ -1;

              }
              else
              {
                 pageBottom_ = 0;
                 cursorRow_  = -1; // skip to input region
              }

              getWindowSize(vp, cmdType_); // and reset screen text attributes
              paintWindow(vp);
              break;



          case CursorWindow::func_tab:
              {
                  char buffer[200];
              
                  // perform tab expansion on the text in the input window based on the most recently executed
                  // process that is reading/writing to the terminal
              
                  std::list<int> users = getProcesses();  // get a list of using processes
              
                  if(users.empty())
                  {
                     vp->beep();
                     return false;
                  }
              
                  int mostRecent = users.back();
              
                  snprintf(buffer, sizeof(buffer), "/proc/%d/cwd", mostRecent);  // prepare to readlink on the current working directory of
                                                                                 // the child process
                  FileName tmp(buffer);
              
                  FileName childCurDir = tmp.readlink();
              
                  if(!childCurDir.empty())
                  {
                      std::string inputLine   =  inputString();
                      size_t      inputIndex  =  inputOffset();
              
                      if(    inputIndex > inputLine.size() 
                             || ( inputIndex != inputLine.size() && inputLine[inputIndex] != ' ' )
                        )
                      {
                         vp->beep();
                         return false;
                      }
              
                      std::string curWord;
                      size_t curWord_start=0;
              
                      {
                         // scan backwards from the current cursor position to the beginning of the current word
              
                         size_t scan = inputIndex;
              
                         if(scan == inputLine.size() )
                           --scan;
              
                         while(scan > 0 && inputLine[scan] == ' ')
                            --scan;
              
                         // scan pts to end character in current word
              
                         while(scan > 0 && inputLine[scan] != ' ')
                            --scan;
              
                         ++scan; // step forward from the blank you just found
              
                         // scan pts to beginning of current word and scan != beginning of input --
                         // we don't do 
              
                         curWord_start = scan;
              
                         while(scan < inputLine.size() && inputLine[scan] != ' ')
                            curWord += inputLine[scan++];
                      }
              
                      if(!curWord.empty())
                      {
                          if(curWord[0] == '/')
                          {
                             vp->beep();
                             return false;
                          }
              
                          FileName tmp(childCurDir);
              
                          tmp += '/';
                          tmp += curWord;
                          tmp += '*';
              
                          FileName::sorted_names_t  matches;
              
                          FileName::find_matching(tmp.c_str(), &matches);
              
                          if(matches.empty())
                          {
                             vp->beep();
                             return false;
                          }

                          // matches now contains all matching substrings

                          std::string replacement;


                          if(matches.size() == 1)
                          {
                             replacement = *matches.begin();
                          }
                          else
                          {
                              CursorWindow::Selection selection("pick one of the following");
                       
                              FileName::sorted_names_t::iterator first = matches.begin(),
                                                                 last  = matches.end();
                                                       
                              while(first != last)
                              {
                                FileName const& cur = *first++;
                               
                                selection += cur;
                               
                              }
                       
                              suppressRepaint_ = true;                      // see operator()()
                              bool die =selection.popup(vp->window());
                              suppressRepaint_ = false;

                              if(die)
                                break;
                               
                              replacement = selection.selection_;
                          }

                          size_t lastCurWordSlash = curWord.find_last_of('/');

                          if(lastCurWordSlash != curWord.npos)
                          {
                              // the user is tabbing on a sub-directory reference -- put the subdir name
                              // on the replacement

                              std::string tmp = curWord.substr(0, lastCurWordSlash+1);

                              replacement = tmp + replacement;

                          }


                          inputLine.erase(curWord_start, curWord.size());
                          inputLine.insert(curWord_start, replacement);
              
                          inputIndex = curWord_start + replacement.size();
              
                      }
              
                      setInputString(vp, inputLine, inputIndex);
                  }
              
                  return false;
              }
              break;


          case CursorWindow::func_enter:
              
              if(cursorRow_ == size_t(-1))
              {
                  
                  inputIndex_ = -1;   // new commands reset position used by ^U/^L to scroll though command list
                  
                  if(   !inputLine_.empty() 
                     && (   input_.empty() 
                         || (    (input_[0] != inputLine_)
                             &&  (inputLine_[0] != '!')
                            )
                        )
                    )
                      input_.push_front(inputLine_);   // don't record empty or !replay lines
                  
                  inputLine_ += '\n';
                  ::write(masterFD_, inputLine_.data(), inputLine_.size());
                  inputLine_.resize(0);
                  inputRow_ = 0;
                  inputCol_ = leftInputCol_;
                  inputOffset_ = 0;
              }
              else
              if(cursorRow_ < output_.size())
              {
                  // the cursor is in the scrollable region

                  inputLine_ = output_[cursorRow_];

                  inputRow_ = 0;
                  inputCol_ = leftInputCol_;
                  inputOffset_ = 0;

                  cursorRow_ = -1;
              }


              paintWindow(vp);

              break;

          case CursorWindow::func_paste:   // paste into command region
              {
                  paste_buffer_type *buffer = paste_buffer();

                  if(!buffer || buffer->empty())
                     break;

                  inputIndex_ = 0;

                  std::string expanded = StrTool::expand_tabs(*buffer->begin());

                  inputLine_ = expanded;


                  inputRow_ = 0;
                  inputCol_ = leftInputCol_;
                  inputOffset_ = 0;
                  
                  paintWindow(vp);

              }
              break;

          case CursorWindow::func_mark:   // mark a section of the scrollable region
             if(cursorRow_ < output_.size())
             {
                // the cursor is not -1 and is less than the size of the scrollable region

                if(markBottom_ == size_t(-1))
                {
                   // there is no mark currently set

                   markBottom_ = markTop_ = cursorRow_;

                }
                else
                {
                   if(cursorRow_ < markBottom_)
                   {
                       // below the bottom
                       markBottom_ = cursorRow_;
                   }
                   else
                   if(cursorRow_ > markTop_)
                   {
                       // above the top
                       markTop_ = cursorRow_;
                   }
                   else
                   {
                       markTop_ = markBottom_ = -1;
                   }

                }

                paintWindow(vp);


             }
             break;

          case CursorWindow::func_cb:  // copy marked block
              if(!output_.empty() && markBottom_ != size_t(-1))
              {
                  // there is output and there is a marked block in it

                  Viewer::paste_buffer_type &buffer = *paste_buffer();

                  buffer.erase(buffer.begin(), buffer.end());


                  for(size_t i = markBottom_;  i <= markTop_; ++i)
                  {
                     buffer.push_back(output_[i]);
                  }


                  markTop_ = markBottom_ = -1;

                  paintWindow(vp);


              }
              break;

          case CursorWindow::func_upword:   // scroll back through previously executed commands
              if(!input_.empty())
              {
                  if(inputIndex_ == size_t(-1))
                  {
                      inputIndex_ = 0;
                      inputLine_ = input_[0];
                  }
                  else
                  if(inputIndex_ < input_.size()-1)
                  {
                      // there is an earlier location to access

                      ++inputIndex_;

                      inputLine_ = input_[inputIndex_];

                  }
                  
                  paintCommandLine(vp);

              }
              break;

          case CursorWindow::func_dnword:   // scroll back through previously executed commands
              if(!input_.empty())
              {
                  if(inputIndex_ == 0)
                  {
                      inputIndex_ = -1;
                      inputLine_ = "";
                  }
                  else
                    if(int(inputIndex_) > 0)
                  {
                      // there is an earlier location to access

                      --inputIndex_;

                      inputLine_ = input_[inputIndex_];

                  }
                  
                  paintCommandLine(vp);

              }
              break;

        

          case CursorWindow::func_dc_prev:
              if(    cursorRow_ == size_t(-1)
                 &&  !inputLine_.empty()
                 &&  inputOffset_ > 0   // implies inputCol_ > leftInputCol_
                )
              {
                 // delete prev char
                 --inputOffset_;
                 inputLine_.erase(inputLine_.begin()+inputOffset_);

                 --inputCol_;

                 if(inputCol_ < leftInputCol_)
                 {
                    inputCol_ = windowCols_-1;

                    --inputRow_;  // if inputOffset_ > 0, this will
                                  // never walk us backwards past
                                  // the first line of the input region
                 }

                 paintCommandLine(vp);


              }
              break;

          case CursorWindow::func_dc:
              if(    cursorRow_ == size_t(-1)
                 &&  !inputLine_.empty()
                )
              {
                 // delete current char

                 if(size_t(inputOffset_) < inputLine_.size())
                 {
                     inputLine_.erase(inputLine_.begin()+inputOffset_);
                     paintCommandLine(vp);
                 }

              }
              break;

          default:

              if(e->value_ == 0x018)
              {
                 ctrlX_ = true;
              }
              else
                if(e->value_ == 0x03)
                  {
                    char ctrlC = 0x03;
                    int rv = write(masterFD_, &ctrlC, 1);

                    if(rv < 0)
                    {
                        char buffer[40];
                        snprintf(buffer, sizeof(buffer), "i/o error: %d", errno);
                        output_.push_front(buffer);
                    }
                  }

              break; // ignore unknown functions
        }
    }
    else
    if(e->type_ == CursorWindow::input_event::DataKey)
    {
        if(cursorRow_ != size_t(-1))
        {
           cursorRow_ = -1;
           pageBottom_= 0;
           getWindowSize(vp, cmdType_);
           paintWindow(vp);
        }


        switch(e->value_)
        {
            case 0x04:
                {
                    static char ctrlD[2] = { 0x04, '\n'};

                    write(masterFD_, &ctrlD[0], 2);

                }
                break;

            case 0x18:  // ^X
               ctrlX_ = true;
               break;

            default:

                if(e->value_ >= ' ' && e->value_ <= '~')
                {
                   if(size_t(inputOffset_) >= inputLine_.size())
                   {
                       inputLine_ += char(e->value_);
                   }
                   else
                   {
                      inputLine_.insert( inputLine_.begin() + inputOffset_, char(e->value_));
                   }

                   ++inputCol_;                      // inputCol_ is about the screen
                   ++inputOffset_;                   // inputOffset_ is an index into inputLine_

                   if(inputCol_ - leftInputCol_ >= availInputCols_)
                   {
                      inputCol_ = leftInputCol_;
                      ++inputRow_;

                      if(inputRow_ == inputRows_)
                      {
                          ++inputRows_;
                          getWindowSize(vp, cmdType_);
                      }
                   }
 
                   paintWindow(vp);


                }

                break;
        }
    }

 
   if(e->value_ != 0x018)
      ctrlX_ = false;


    return false;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

static std::string appname("DumbTerminal");

std::string const &
DumbTerminalViewer::
application_name() const
{
   return appname;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

std::string 
DumbTerminalViewer::
description() const
{
    return "A viewer that lets the user issue commands to a dumb terminal program";
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

void 
DumbTerminalViewer::
help()
{

  std::auto_ptr< std::list< std::string > >
    viewer_help( new std::list< std::string > );

  viewer_help->push_back("");
  viewer_help->push_back("");
  viewer_help->push_back("Help for the Terminal Viewer");
  viewer_help->push_back("");
  viewer_help->push_back("  Overview");
  viewer_help->push_back("");
  viewer_help->push_back("    A terminal viewer exists so that you can interact with another program.");
  viewer_help->push_back("    The terminal viewer only supports dumb terminal behavior, so  you can't");
  viewer_help->push_back("    run VI, EMACS, or any other curses aware program in the window.  It is");
  viewer_help->push_back("    mean mainly so that you can issue command line commands and see their output.");
  viewer_help->push_back("    You should be able to log in, however.");
  viewer_help->push_back("");
  viewer_help->push_back("    The Terminal Viewer page has 3 regions:  title, main window, and input.");
  viewer_help->push_back("");
  viewer_help->push_back("    The input area is editable, it appears at the bottom with different");
  viewer_help->push_back("    text attributes than the main window");
  viewer_help->push_back("");
  viewer_help->push_back("    The main window shows a history of what has gone on in the terminal.");
  viewer_help->push_back("    Note that when interacting with a shell, such as bash, the command prompt");
  viewer_help->push_back("    appears above the input region, not in it.");
  viewer_help->push_back("");
  viewer_help->push_back(" Keyboard");
  viewer_help->push_back("");
  viewer_help->push_back("    Most keys work as normal.");
  viewer_help->push_back("");
  viewer_help->push_back("    ^U  Scrolls the input window backwards through the list of commands executed");
  viewer_help->push_back("        so that you can select one for re-execution.");
  viewer_help->push_back("");
  viewer_help->push_back("    ^L  Lets you scroll forward in the command list brought up by ^U.");
  viewer_help->push_back("");
  viewer_help->push_back("    ^X ^U  Lets you increase the height of the input window.");
  viewer_help->push_back("    ^X ^Y  Lets you decrease the height of the input window -- if allowed.");



  manager_->help_helper(*viewer_help);
    
}
                        
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

bool 
DumbTerminalViewer::
is_dirty() const
{
    return false;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

Viewer*
DumbTerminalViewer::
app(ViewerManager*vm, std::string &fullname)
{
   int masterFD=-1;

   int rc = forkpty(&masterFD, 0, 0, 0);

   if(rc == -1)
   {
      masterFD = -1;
   }
   else
   if(rc == 0)
   {
      exit(1);    // child process won't be doing anything
   }


   DumbTerminalViewer *rv =new DumbTerminalViewer(vm, masterFD);  

   rv->title_ = fullname;

   return rv;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

void 
DumbTerminalViewer::
getWindowSize(CursorWindow::viewport *vp, int cmd)
{
    // get windowRows_, windowCols_, windowOffset_

    CursorWindow::row_col size = vp->size();
    CursorWindow::row_col orig = vp->origin();

    bool widthChanged = false;

    if(windowCols_ != size.col_)
       widthChanged = true;

    windowRows_   = size.row_;
    windowCols_   = size.col_;
    windowOffset_ = orig.col_; 
    cmdType_      = cmd;

    leftInputCol_ = windowOffset_ ? 1 : 0;
    availInputCols_ =windowCols_ - leftInputCol_;

    scrollRows_   = windowRows_ - inputRows_ - titleRows_;

    std::string inputText;

    int neededRows = inputArea(inputText);

    if(widthChanged || (inputRow_ >= neededRows))
    {
        inputRow_ = neededRows - 1;

        inputCol_ = (inputText.size() % availInputCols_) + leftInputCol_;

        inputOffset_ = inputText.size();

    }


    switch(cmd)
    {
       case   CursorWindow::viewport::repaint_handler::deactivate:
           // inactive
           {
               if(cursorRow_ != size_t(-1))
               {
                  // the cursor is in the scrolling region, not the input
                  // region.
               
                  inputAtt_  = ViewerManager::normal_att;
               }
               else
               {
                   // inputAtt_ = ViewerManager::active_title_att;     // the cursor is in the input region
               
                   inputAtt_ = ViewerManager::normal_att;     // the cursor is in the input region
               }
               
               
               bottomAtt_= ViewerManager::bottom_att;             // text attribute of last line of output_ displayed
               outputAtt_= ViewerManager::normal_att;             // text attribute of most of the output_ region
               titleAtt_ = ViewerManager::inactive_title_att;     // text attribute of top line of window
               cursorAtt_= CursorWindow::bold_ul;                 // text attribute of cursor line in the output_
               markedAtt_= CursorWindow::bold;                    // text attribute of output_ regions which have been marked
               markedCursorAtt_ = CursorWindow::bold_ul;          // only used if the currentRow_ is the cursor and the line is marked
               botMarkCursorAtt_ = CursorWindow::underlined;
               botCursorAtt_ = CursorWindow::bold_ul;
           }
           break;

       default:
           {
               if(cursorRow_ != size_t(-1))
               {
                  // the cursor is in the scrolling region, not the input
                  // region.
               
                  inputAtt_  = ViewerManager::normal_att;
               }
               else
               {
                   // inputAtt_ = ViewerManager::active_title_att;     // the cursor is in the input region
               
                   inputAtt_ = ViewerManager::normal_att;     // the cursor is in the input region
               }
               
               
               bottomAtt_= ViewerManager::bottom_att;           // text attribute of last line of output_ displayed
               outputAtt_= ViewerManager::normal_att;           // text attribute of most of the output_ region
               titleAtt_ = ViewerManager::active_title_att;     // text attribute of top line of window
               cursorAtt_= ViewerManager::highlighted_att;      // text attribute of cursor line in the output_
               markedAtt_= CursorWindow::reversed;              // text attribute of output_ regions which have been marked
               markedCursorAtt_ = CursorWindow::reverse_ul;    // only used if the currentRow_ is the cursor and the line is marked
               botMarkCursorAtt_ = CursorWindow::rev_ul_bold;
               botCursorAtt_ = CursorWindow::reverse_ul;
           }
           break;
    }

}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
void
DumbTerminalViewer::
paintTitle(CursorWindow::viewport *vp)
{
   vp->set_text_attribute(titleAtt_);
   vp->set_curpos(0,0);

   if(windowOffset_)
   {
      vp->write("|");
   }

   vp->write(title_);
   vp->write("    F1=help");
   vp->fill_to_eol();


}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
void
DumbTerminalViewer::
paintWindow(CursorWindow::viewport *vp)
{
   paintTitle(vp);
   paintOutput(vp);
   paintCommandLine(vp);
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
void
DumbTerminalViewer::
paintOutput(CursorWindow::viewport *vp)
{
    for(int row = 0; row < scrollRows_; ++row)
    {
        paintRow(vp, row);
    }

}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
void
DumbTerminalViewer::
paintRow(CursorWindow::viewport *vp, int row)
{
    // paint a row in the scrolling region

    size_t rowOffset = pageBottom_ + row;

    if(row == 0)
    {
        if(cursorRow_ == pageBottom_)
        {
            vp->set_text_attribute(isMarked(rowOffset) ? botMarkCursorAtt_ : markedAtt_);
        }
        else
            vp->set_text_attribute(isMarked(rowOffset) ? botCursorAtt_ : bottomAtt_);
    }
    else
    {
        if(rowOffset == cursorRow_)
        {
            vp->set_text_attribute(isMarked(rowOffset) ? markedCursorAtt_ : cursorAtt_);
        }
        else
            vp->set_text_attribute(isMarked(rowOffset) ? markedAtt_ : outputAtt_);
    }

    vp->set_curpos(windowRows_ - 1 - inputRows_ - row, 0);

    int leftBorder=0;

    if(windowOffset_)
    {
       leftBorder=1;
       vp->write("|");
    }


    if(rowOffset < output_.size())
    {
        std::string data = output_[rowOffset];

        size_t availableColumns = windowCols_ - leftBorder;

        size_t neededColumns = data.size() - leftOffset_;

        if(neededColumns > availableColumns)
           neededColumns = availableColumns;

        if(leftOffset_ < data.size())
            vp->write(data.data() + leftOffset_, neededColumns);

    }

    vp->fill_to_eol();


}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
void
DumbTerminalViewer::
paintCommandLine(CursorWindow::viewport *vp)
{
    vp->set_text_attribute(inputAtt_);

    std::string expanded = StrTool::expand_tabs(inputLine_);

    int neededRows = inputArea();

    if(neededRows > inputRows_)
    {
        if(inputRows_ < windowRows_ - titleRows_ - neededRows)
            inputRows_ = neededRows;
    }

    char const *p = expanded.data();

    size_t      remaining = expanded.size();

    for(int i = 0; i < inputRows_; ++i)
    {
       vp->set_curpos(windowRows_ - inputRows_ + i, 0);

       if(windowOffset_)
          vp->write('|');

       if(remaining > size_t(availInputCols_))
       {
           vp->write(p, availInputCols_);
           p += availInputCols_;
           remaining -= availInputCols_;
       }
       else
       if(remaining)
       {
          vp->write(p, remaining);
          remaining = 0;
       }


       vp->fill_to_eol();

    }

    if(cursorRow_ == size_t(-1))
        vp->set_curpos(windowRows_ - inputRows_ + inputRow_, inputCol_);
    else
        vp->set_curpos(windowRows_ - inputRows_ + inputRow_, windowCols_-1);



}



///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
DumbTerminalViewer::
DumbTerminalViewer(ViewerManager* vm, int masterFD)
:  Viewer(vm)
,  pageBottom_(0)
,  titleRows_(1)
,  inputRows_(1)
,  inputRow_(0)
,  inputOffset_(0)
,  availInputCols_(0)
,  leftInputCol_(0)
,  inputCol_(0)
,  windowRows_(0)
,  windowCols_(0)
,  windowOffset_(0)
,  scrollRows_(0)
,  leftOffset_(0)
,  cursorRow_(-1)
,  cmdType_(0)
,  inputAtt_(ViewerManager::normal_att)
,  bottomAtt_(ViewerManager::normal_att)
,  outputAtt_(ViewerManager::normal_att)
,  titleAtt_(ViewerManager::normal_att)
,  cursorAtt_(ViewerManager::normal_att)
,  markedAtt_(ViewerManager::normal_att)
,  ctrlX_(0) 
,  thread_(0)
,  shuttingDown_(0)
,  dataArrived_(false)
,  inputIndex_(-1)   // -1 is the "not scrolling" flag
,  markTop_(-1)
,  markBottom_(-1)
,  markedCursorAtt_(ViewerManager::normal_att)
,  botMarkCursorAtt_(ViewerManager::normal_att)
,  botCursorAtt_(ViewerManager::normal_att)
,  suppressRepaint_(false)  // see handle_event() processing of func_tab
{

    if(masterFD != -1)
    {
        ptyGood_ = true;
        masterFD_ = masterFD;

        TerminalStatus status(masterFD_);

        char buffer[256];

        buffer[0]=0;

        ptsname_r(masterFD_, buffer, sizeof(buffer));

        buffer[sizeof(buffer)-1] = 0;

        terminalName_ = buffer;

        status.setNoEcho();
        status.setSignal(true);

        status.setStatus(masterFD_);

        output_.push_front(std::string("Using terminal:  ") + ptyName());

        thread_ = createThread(&DumbTerminalViewer::reader, this, true); // create a joinable thread

    }
    else
    {
       ptyGood_ = false;
       output_.push_front("Error:  could not open terminal or launch child process");
    }


}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
int
DumbTerminalViewer::
inputArea(std::string &expanded) const
{
    expanded.resize(0);

    expanded = StrTool::expand_tabs(inputLine_);

    int rows = expanded.size() + availInputCols_;

    rows /= availInputCols_;


    if(rows < 1)
       rows = 1;

    return rows;
    
}

void* 
DumbTerminalViewer::
reader(void* that)
{
    DumbTerminalViewer* me =  reinterpret_cast<DumbTerminalViewer*>(that);

    // std::string terminalOutput;

    int fh = me->masterFD_;

    bool lineStart=true;
    int  lineCol=0;

    for(;;)
    {
        char c=0;

        errno = 0;

        int rv =  read(fh, &c, 1);

        int errnoValue = errno;

        if(errnoValue == EIO)
        {
           // some kind of general error -- perhaps there is nothing reading
           // from the terminal

           usleep(10000);  // sleep 10 miliseconds so we don't burn up too much cpu

        }
        else
        if(errnoValue == EINTR)
        {
           // a signal interrupted the io

           break;
        }


        if(rv < 1)
        {
           // end of file has been read, handle any partial output buffers remaining


           // check to see if the user has shutdown the window or if some other eve3n
           // has caused us to read eof

           {
               ScopedMutexLock lock(me->shutDownMutex_);

               if(me->shuttingDown_)
                   break;  // exit the thread because the user has exited this page.

               // otherwise, we are not going exit this loop and we will assume
               // that one program has closed the terminal as an output device.  Others
               // may being writing to it.

           }
        }
        else
        if(c == '\n')
        {
            lineStart = true;
            lineCol   = 0;

            {
               ScopedMutexLock lock(me->dataArrivedMutex_);
            
               me->dataArrived_=true;
            }

        }
        else
        if(c != '\r')
        {
           if(lineStart)
           {
              me->outputLine("");
              lineStart = false;
           }

           if(c != '\t')
           {
               me->appendOutput(c);
               ++lineCol;
           }
           else
           {
              // handling tabls

              int nextTab = lineCol / 8;
              ++nextTab;
              nextTab *= 8;

              int spaces = nextTab - lineCol;

              while(spaces--)
              {
                 me->appendOutput(' ');
                 ++lineCol;
              }



           }

            {
               ScopedMutexLock lock(me->dataArrivedMutex_);
            
               me->dataArrived_=true;
            }


        }

    }


    return 0;
}

void
DumbTerminalViewer::
outputLine(std::string const &line)
{
   ScopedMutexLock lock(outputMutex_);

   if(!line.empty() && line[ line.size()-1 ] == '\r')
   {
      std::string copy(line);

      copy.erase(copy.size()-1);
      output_.push_front(copy);

   }
   else
     output_.push_front(line);



}

void
DumbTerminalViewer::
appendOutput(char c)
{
   ScopedMutexLock lock(outputMutex_);

   if(output_.empty())
      output_.push_front("");

   output_[0]+= c;



}

void
DumbTerminalViewer::
terminate()
{
     ScopedMutexLock lock(shutDownMutex_);
     shuttingDown_ = true;
}

std::list<int> 
DumbTerminalViewer::
getProcesses()
{
    std::string cmd("lsof -t " + terminalName_ + " </dev/null 2>&1 | cat");  // launch lsof without it being in the list of processes that it returns

    FILE *processList = popen(cmd.c_str(), "r");

    std::list<int> rv;

    if(processList)
    {
        char buffer[2000];

        while(fgets(buffer, sizeof(buffer), processList))
        {
           int v=0;

           sscanf(buffer, "%d", &v);

           if(v != 0)
              rv.push_back(v);

        }

        pclose(processList);
    }
    return rv;

}



} // namespace cxxtls

#else
   int no_can_do_on_windows_big_guy;
#endif
