//
// Copyright 2013, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include <portable_io.h>
#include <cxxtls/foreach.h>
#include <cxxtls/varobj.h>
#include <cxxtls/strtool.h>
#include <stdio.h>
#include <ctype.h>
#include <memory>
#include <fstream>

namespace cxxtls
{

typedef VarObj::ParseResult  ParseResult;

void skipLeadingSpaces(char const *&f, char const *&l)
    /// Modify f and l to skip leading spaces.
{
   while(f !=l)
   {
      char c = *f;

      if(!isspace(c))
         break;

      ++f;

   }
}

inline bool isFloatDigit(char c)
{
    return    isdigit(c)
           || c == '-'
           || c == '+'
           || c == '.'
           || c == 'e'
           || c == 'E'
           ;
}


std::string prepareError(std::string const &msg, char const *offStart, char const *offEnd)
    // given an error message and range of characters,
    // construct an error message that shows the leading part of the error
    // string defined by the range.
{
   static std::string error("Error:  ");

   std::string shortText;

   if(offEnd - offStart > 20)
   {
       shortText.assign(offStart, offStart + 20);
   }
   else
   {
       shortText.assign(offStart, offEnd);
   }


   return error + msg + " " + shortText;
}



//=======================================================================
//=======================================================================
//=======================================================================
VarObj::~VarObj()
{
}

//=======================================================================
//=======================================================================
//=======================================================================
VarObj*  
VarObj::
findChild(std::string const &)
{
     return 0;
}

//=======================================================================
//=======================================================================
//=======================================================================
bool 
VarObj::
leaf() const
{
   return 1;
}
//=======================================================================
//=======================================================================
//=======================================================================
std::string
VarObj::
parsable_value() const
{
    return value();
}

//=======================================================================
//=======================================================================
//=======================================================================
std::string
StringVarObj::
value() const
{
    return value_;
}

//=======================================================================
//=======================================================================
//=======================================================================


std::string
VarObj::
encodeString(std::string const &value)
{
   std::string rv;

   rv += '"';

   CXXTLS_FOREACH(char c, value)
   {
       switch(c)
       {
            case '\t': rv += "\\t"; break; // tab
            case '\n': rv += "\\n"; break; // newline
            case '\r': rv += "\\r"; break; // carriage return
            case '\b': rv += "\\b"; break; // backspace
            case '\a': rv += "\\a"; break; // bel
#ifdef _MSC_VER
            case 0x1b: rv += "\\e"; break; // escape
#else
            case '\e': rv += "\\e"; break; // escape
#endif
            case '"' : rv += "\\\"";break;

           default:  rv += c;
                     break;
       }
   }

   rv += '"';

   return rv;
}


//=======================================================================
//=======================================================================
//=======================================================================
std::string
StringVarObj::
parsable_value() const
{
   return VarObj::encodeString(value_);
}


//=======================================================================
//=======================================================================
//=======================================================================
std::string
IntegerVarObj::
value() const
{
    char buffer[100];

    snprintf(buffer, 100, "%ld", value_);

    return buffer;
}

//=======================================================================
//=======================================================================
//=======================================================================
std::string
FloatVarObj::
value() const
{
    char buffer[1000];

    snprintf(buffer, sizeof(buffer), "%g", value_);

    return buffer;
}

//=======================================================================
//=======================================================================
//=======================================================================
bool
CompositeVarObj::
leaf() const
{
    return false;
}
//=======================================================================
//=======================================================================
//=======================================================================
VarObj*  
CompositeVarObj::
findChild(std::string const &name)
{
    ValueType::iterator where = value_.find(name);

    if(where == value_.end())
        return 0;

    return where->second;
}
//=======================================================================
//=======================================================================
//=======================================================================
VarObj*  
CompositeVarObj::
release(std::string const &name)
{
    ValueType::iterator where = value_.find(name);

    if(where == value_.end())
        return 0;

    VarObj *rv = where->second;

    value_.erase(where);

    return rv;
}


//=======================================================================
//=======================================================================
//=======================================================================
std::string
CompositeVarObj::
value() const
{
    std::string rv="{";

    typedef ValueType::value_type MapValue;

    ValueType::const_iterator f = value_.begin(),
                              l = value_.end();


    while(f != l)               
    {
       MapValue const &v = *f++;

       rv += v.first;
       rv += '=';
       rv += v.second->parsable_value();

       if(f != l)
          rv += ',';
    }

    rv += "}";

    return rv;
}

//=======================================================================
//=======================================================================
//=======================================================================
CompositeVarObj::
~CompositeVarObj()
{
    CXXTLS_FOREACH(ValueType::value_type &v, value_)
    {
        delete v.second;
    }
}
//=======================================================================
//=======================================================================
//=======================================================================

static
VarObj::ParseResult
parseNumber(char const *&f, char const *&l)
{
    std::string tmp;

    bool        isFloat = false;


    while(  f != l )
    {
        char c = *f;

        if(    !isdigit(c) 
           &&  (c != '-')
          )
        {
          isFloat = isFloatDigit(c);
        }

        if(!isFloatDigit(c))
           break;

        tmp += c;

        ++f;
    }

    if(tmp.empty())
        return ParseResult(prepareError("Unexpected numeric format:  ", f, l) );

    if(isFloat)
    {
        double d=0;

        if(!sscanf(tmp.c_str(), "%lf", &d))
            return ParseResult(prepareError("Invalid float format:  ", 
                                            tmp.data(),
                                            tmp.data() + tmp.size()
                                           )
                              );

        return ParseResult(new FloatVarObj(d) );
    }


    long d=0;

    if(!sscanf(tmp.c_str(), "%ld", &d))
        return ParseResult(prepareError("Invalid integer format:  ", 
                                        tmp.data(),
                                        tmp.data() + tmp.size()
                                       )
                          );

    return ParseResult(new IntegerVarObj(d) );

}

//=======================================================================
//=======================================================================
//=======================================================================
VarObj::ParseResult
parseString(char const *&f, char const *&l)
{
    // parse something like this:
    //
    //    "soemthing\notherthing"

    std::string rv;

    skipLeadingSpaces(f,l);

    if(f == l)
         return ParseResult("Error:  Empty String Input");


    if(*f != '"')
        return ParseResult("Error;  Expected leading '\"'");

    ++f;

    while(f != l)
    {
        char c = *f;

        if( c == '"')
          {

            break;
          }

        if(c == '\\')
        {
           ++f;

           if(f == l)
              break;

           switch(*f)
           {
              case 't': c = '\t'; break; // tab
              case 'n': c = '\n'; break; // newline
              case 'r': c = '\r'; break; // carriage return
              case 'b': c = '\b'; break; // backspace
              case 'a': c = '\a'; break; // bel
              case 'e': c = 0x1b; break; // escape

              default:  c = *f;
                        break;
           }
        }

        rv += c;
        ++f;
    }

    if( f == l )
        return ParseResult(prepareError("Expected trailing '\"':  ", f, l));

    ++f; // skip trailing "

    return ParseResult(new StringVarObj(rv) );
}
//=======================================================================
//=======================================================================
//=======================================================================
bool
parseName(std::string &name, char const *&f, char const *&l)
{
    //
    // consume a name from the input stream (f->l).
    //
    // return true if successful.
    //

    skipLeadingSpaces(f, l);

    if(f == l)
        return false;

    if(!isalnum(*f) && *f != '_')
       return false;

    while(f != l)
    {
       char c = *f;

       if(!isalnum(c) && !isdigit(c) && (c != '_') && (c != '-') )
         break;

       name += c;
       ++f;
    }



    return true;
}
//=======================================================================
//=======================================================================
//=======================================================================
VarObj::ParseResult
VarObj::
parseList(char const *&f, char const *&l)
{
    // Parse a comma separated list of name=value pairs where the
    // values are VarObjs

    skipLeadingSpaces(f,l);

    if(f == l)
        return ParseResult("Error:  empty list input");

    std::auto_ptr<CompositeVarObj>  rv(new CompositeVarObj());

    while(f != l)
    {

       skipLeadingSpaces(f,l);

       if(*f == '}')
          break;

       std::string name;

       if(!parseName(name, f, l))
          break;

       if(f == l)
          break;

       skipLeadingSpaces(f,l);

       if(*f != '=')
       {
          return ParseResult(
                               prepareError("Expected =<some value>", f, l)
                            );
       }

       ++f;

       skipLeadingSpaces(f,l);

       if(f == l)
          break;

       ParseResult tmp = VarObj::parse(f, l);

       if(tmp.rv_ == 0)
       {
          return tmp;
       }

       rv->adopt(name, tmp.rv_);

       if(f != l && *f == ',')
          ++f;


        
    }

    if(f !=l && *f != '}')
    {
         return ParseResult(  prepareError("Expected name=value or trailing '}' or end of string",
                                           f,
                                           l
                                          )
                           );
    }



    if(rv->empty())
    {
        return ParseResult("Error:  empty list");
    }

    return ParseResult(rv.release());
}



//=======================================================================
//=======================================================================
//=======================================================================
VarObj::ParseResult
parseComposite(char const *&f, char const *&l)
{
   skipLeadingSpaces(f,l);

   if(f == l)
        return ParseResult("Error:  empty composite object input");

   if(*f != '{')
   {
       return ParseResult("Error:  expected leading '{'");
   }

   ++f;

   ParseResult rv = VarObj::parseList(f, l);

   if(rv.rv_ == 0)
       return rv;

   // we got a successful parse, but:

   if(f == l || *f != '}')
   {
      delete rv.rv_;

      return ParseResult("Error:  expected trailing '}'");
   }

   ++f;

   return rv;

}

//=======================================================================
//=======================================================================
//=======================================================================
VarObj::ParseResult
parseArray(char const *&f, char const *&l)
{
   skipLeadingSpaces(f,l);

   if(f == l)
        return ParseResult("Error:  empty array input");

   if(*f != '[')
   {
       return ParseResult("Error:  expected leading '['");
   }

   ++f;

   skipLeadingSpaces(f,l);

   ArrayVarObj * rv = new ArrayVarObj();

   while(    (f != l)
         &&  (*f != ']')
        )
   {
       skipLeadingSpaces(f,l);

       ParseResult tmp = VarObj::parse(f,l);
       
       if(tmp.rv_ == 0)
       {
           delete rv;
           return tmp;
       }

       rv->adopt(tmp.rv_);

       skipLeadingSpaces(f,l);

       if(f != l && *f == ',')
           ++f;
   }

   if(f == l || *f != ']')
   {
       delete rv;

       return ParseResult(prepareError("Expected array end ]", f, l));
   }

   ++f;

   return ParseResult(rv);

}

//=======================================================================
//=======================================================================
//=======================================================================
VarObj::ParseResult
VarObj::
parse(char const *&f, char const *&l)
{
    skipLeadingSpaces(f,l);

    if(f == l)
    {
       return ParseResult("empty input");
    }

    switch(*f)
    {
         case '0': case '1': case '2': case '3': case '4': case '5':
         case '6': case '7': case '8': case '9':case '-':  

         {
              return parseNumber(f,l);
         }
         break;

         case '{':  
         {
             return parseComposite(f, l);
         }
         break;

         case '"':
         {
            return parseString(f,l);
         }
         break;

         case '[':
         {
             return parseArray(f,l);
         }
         break;

         default:
           if(isalpha(*f) || *f == '-' || *f == '_' )
           {
             // parse name=value and call it a composite

             std::auto_ptr<CompositeVarObj>  rv(new CompositeVarObj);

             std::string name;

             char const *saved = f;

             if(!parseName(name, f, l))
               {
                 // huh?  we have something like this:   [fred]

                 f = saved;

                 return ParseResult(prepareError("Expected name=value", f, l));
               }

                   ++f; // skip =

                 ParseResult tmp = parse(f,l);

                 if(tmp.rv_)
                   {
                     // got a successful parse

                     rv->adopt(name, tmp.rv_);

                     return ParseResult(rv.release());
                   }
                 else
                   return ParseResult(tmp.error_);
             
           }
           break;

    }

    return ParseResult(prepareError("Unknown token type:  ", f, l) );

}

//=======================================================================
//=======================================================================
//=======================================================================
ArrayVarObj::~ArrayVarObj()
{
    CXXTLS_FOREACH( VarObj *cur, value_)
    {
        delete cur;
    }
}

//=======================================================================
//=======================================================================
//=======================================================================
bool 
ArrayVarObj::
leaf() const
{
   return 0;
}
//=======================================================================
//=======================================================================
//=======================================================================
std::string
ArrayVarObj::
value() const
{
   std::string rv("[");

   ValueType::const_iterator f = value_.begin(),
                             l = value_.end();
   
   while(f != l)
   {
       VarObj const *cur = *f++;

       rv += cur->value();

       if(f !=l)
           rv += ',';

   }


   rv += "]";

   return rv;
}

//=======================================================================
//=======================================================================
//=======================================================================
VarObj*
ArrayVarObj::operator[](size_t i)
{
   if(i < value_.size())
      return value_[i];

   return 0;
    
}

//=======================================================================
//=======================================================================
//=======================================================================
VarObj const *
ArrayVarObj::operator[](size_t i) const
{
   if(i < value_.size())
      return value_[i];

   return 0;
    
}

//=======================================================================
//=======================================================================
//=======================================================================
VarObj *
VarObj::findNestedChild(std::string const &path)
{
    StrTool::stringlist_t words;

    StrTool::parse_words(path, &words, 100000, StrTool::Is_Delim('/'));

    VarObj *scan = this;  // scan is the "parent" until the leave is found

    CXXTLS_FOREACH(std::string const &word, words)
    {
        VarObj *next = scan->findChild(word);

        if(!next)
        {
           scan = 0;
           break;   // no can find big guy!
        }

        scan = next;

        // when we are out of words, scan points to the most recently found
        // word's VarObj, so we have found the one we are looking for

    }




    return scan;

    
}

//=======================================================================
//=======================================================================
//=======================================================================
VarObj const *
VarObj::findNestedChild(std::string const &path) const
{
   VarObj *me = const_cast<VarObj*>(this);

   return me->findNestedChild(path);

}
//=======================================================================
//=======================================================================
//=======================================================================
ParseResult 
VarObj::
readAndParseFile(std::string const &filename)
{
    std::string contents;

    std::ifstream infile(filename.c_str());

    if(!infile.good())
    {
        return ParseResult(std::string("Cannot open file: ") + filename);
    }

    if(std::getline(infile, contents, char(0)))
    {
        char const *f = &*contents.begin();
        char const *l = &*contents.end();

        return parse(f, l);
    }

    return ParseResult(std::string("Error, unable to read file: ") + filename);



}

//=======================================================================
//=======================================================================
//=======================================================================



} // namespace cxxtls
